@@ -48,18 +48,10 @@ static const struct got_error *got_get_repo_commit(struct request *,
     struct got_object_id *);
 static const struct got_error *got_gotweb_dupfd(int *, int *);
 static const struct got_error *got_gotweb_openfile(FILE **, int *, int *);
-static const struct got_error *got_gotweb_flushfile(FILE *, int);
 static const struct got_error *got_gotweb_blame_cb(void *, int, int,
     struct got_commit_object *,struct got_object_id *);
 
-static int
-isbinary(const uint8_t *buf, size_t n)
-{
-	return memchr(buf, '\0', n) != NULL;
-}
-
-
-static const struct got_error *
+const struct got_error *
 got_gotweb_flushfile(FILE *f, int fd)
 {
 	if (fseek(f, 0, SEEK_SET) == -1)
@@ -108,7 +100,7 @@ got_gotweb_dupfd(int *priv_fd, int *fd)
 }
 
 const struct got_error *
-got_get_repo_owner(char **owner, struct request *c, char *dir)
+got_get_repo_owner(char **owner, struct request *c)
 {
 	struct server *srv = c->srv;
 	struct transport *t = c->t;
@@ -134,7 +126,7 @@ got_get_repo_owner(char **owner, struct request *c, char *dir)
 }
 
 const struct got_error *
-got_get_repo_age(char **repo_age, struct request *c, char *dir,
+got_get_repo_age(char **repo_age, struct request *c,
     const char *refname, int ref_tm)
 {
 	const struct got_error *error = NULL;
@@ -572,7 +564,7 @@ got_get_repo_tags(struct request *c, int limit)
 	    repo_dir->name) == -1)
 		return got_error_from_errno("asprintf");
 
-	if (qs->commit == NULL && qs->action == TAGS) {
+	if (qs->commit == NULL && (qs->action == TAGS || qs->action == RSS)) {
 		error = got_ref_open(&ref, repo, qs->headref, 0);
 		if (error)
 			goto err;
@@ -800,8 +792,9 @@ err:
 	return error;
 }
 
-const struct got_error *
-got_output_repo_tree(struct request *c)
+int
+got_output_repo_tree(struct request *c,
+    int (*cb)(struct template *, struct got_tree_entry *))
 {
 	const struct got_error *error = NULL;
 	struct transport *t = c->t;
@@ -812,10 +805,10 @@ got_output_repo_tree(struct request *c)
 	struct got_object_id *tree_id = NULL, *commit_id = NULL;
 	struct got_reflist_head refs;
 	struct got_tree_object *tree = NULL;
+	struct got_tree_entry *te;
 	struct repo_dir *repo_dir = t->repo_dir;
-	const char *name, *folder;
 	char *escaped_name = NULL, *path = NULL;
-	int nentries, i, r;
+	int nentries, i;
 
 	TAILQ_INIT(&refs);
 
@@ -852,106 +845,10 @@ got_output_repo_tree(struct request *c)
 
 	nentries = got_object_tree_get_nentries(tree);
 
-	folder = qs->folder ? qs->folder : "";
-
 	for (i = 0; i < nentries; i++) {
-		const char *modestr;
-		struct got_tree_entry *te;
-		mode_t mode;
-
 		te = got_object_tree_get_entry(tree, i);
-
-		mode = got_tree_entry_get_mode(te);
-		if (got_object_tree_entry_is_submodule(te))
-			modestr = "$";
-		else if (S_ISLNK(mode))
-			modestr = "@";
-		else if (S_ISDIR(mode))
-			modestr = "/";
-		else if (mode & S_IXUSR)
-			modestr = "*";
-		else
-			modestr = "";
-
-		name = got_tree_entry_get_name(te);
-		error = gotweb_escape_html(&escaped_name, name);
-		if (error)
-			goto done;
-
-		if (S_ISDIR(mode)) {
-			struct gotweb_url url = {
-				.index_page = -1,
-				.page = -1,
-				.action = TREE,
-				.commit = rc->commit_id,
-				.path = qs->path,
-				/* `folder' is filled later */
-			};
-			char *path = NULL;
-
-			if (fcgi_printf(c,"<div class='tree_wrapper'>\n"
-			    "<div class='tree_line'>") == -1)
-				goto done;
-
-			if (asprintf(&path, "%s/%s", folder, name) == -1) {
-				error = got_error_from_errno("asprintf");
-				goto done;
-			}
-			url.folder = path;
-			r = gotweb_link(c, &url, "%s%s", escaped_name,
-			    modestr);
-			free(path);
-			if (r == -1)
-				goto done;
-
-			if (fcgi_printf(c, "</div>\n" /* .tree_line */
-			    "<div class='tree_line_blank'>&nbsp;</div>\n"
-			    "</div>\n") == -1)
-				goto done;
-		} else {
-			struct gotweb_url url = {
-				.index_page = -1,
-				.page = -1,
-				.path = qs->path,
-				.commit = rc->commit_id,
-				.folder = folder,
-				.file = name,
-			};
-
-			if (fcgi_printf(c, "<div class='tree_wrapper'>\n"
-			    "<div class='tree_line'>") == -1)
-				goto done;
-
-			url.action = BLOB;
-			r = gotweb_link(c, &url, "%s%s", escaped_name,
-			    modestr);
-			if (r == -1)
-				goto done;
-
-			if (fcgi_printf(c, "</div>\n" /* .tree_line */
-			    "<div class='tree_line_blank'>") == -1)
-				goto done;
-
-			url.action = COMMITS;
-			r = gotweb_link(c, &url, "commits");
-			if (r == -1)
-				goto done;
-
-			if (fcgi_printf(c, " | ") == -1)
-				goto done;
-
-			url.action = BLAME;
-			r = gotweb_link(c, &url, "blame");
-			if (r == -1)
-				goto done;
-
-			if (fcgi_printf(c,
-			    "</div>\n"		/* .tree_line_blank */
-			    "</div>\n") == -1)	/* .tree_wrapper */
-				goto done;
-		}
-		free(escaped_name);
-		escaped_name = NULL;
+		if (cb(c->tp, te) == -1)
+			break;
 	}
 done:
 	free(escaped_name);
@@ -963,11 +860,16 @@ done:
 		got_object_tree_close(tree);
 	free(commit_id);
 	free(tree_id);
-	return error;
+	if (error) {
+		log_warnx("%s: %s", __func__, error->msg);
+		return -1;
+	}
+	return 0;
 }
 
 const struct got_error *
-got_output_file_blob(struct request *c)
+got_open_blob_for_output(struct got_blob_object **blob, int *fd,
+    int *binary, struct request *c)
 {
 	const struct got_error *error = NULL;
 	struct transport *t = c->t;
@@ -976,14 +878,15 @@ got_output_file_blob(struct request *c)
 	struct got_commit_object *commit = NULL;
 	struct got_object_id *commit_id = NULL;
 	struct got_reflist_head refs;
-	struct got_blob_object *blob = NULL;
 	char *path = NULL, *in_repo_path = NULL;
-	int obj_type, set_mime = 0, fd = -1;
-	size_t len, hdrlen;
-	const uint8_t *buf;
+	int obj_type;
 
 	TAILQ_INIT(&refs);
 
+	*blob = NULL;
+	*fd = -1;
+	*binary = 0;
+
 	if (asprintf(&path, "%s%s%s", qs->folder ? qs->folder : "",
 	    qs->folder ? "/" : "", qs->file) == -1) {
 		error = got_error_from_errno("asprintf");
@@ -1021,69 +924,66 @@ got_output_file_blob(struct request *c)
 		goto done;
 	}
 
-	error = got_gotweb_dupfd(&c->priv_fd[BLOB_FD_1], &fd);
+	error = got_gotweb_dupfd(&c->priv_fd[BLOB_FD_1], fd);
 	if (error)
 		goto done;
 
-	error = got_object_open_as_blob(&blob, repo, commit_id, BUF, fd);
+	error = got_object_open_as_blob(blob, repo, commit_id, BUF, *fd);
 	if (error)
 		goto done;
-	hdrlen = got_object_blob_get_hdrlen(blob);
-	do {
-		error = got_object_blob_read_block(&len, blob);
-		if (error)
-			goto done;
-		buf = got_object_blob_get_read_buf(blob);
 
-		/*
-		 * Skip blob object header first time around,
-		 * which also contains a zero byte.
-		 */
-		buf += hdrlen;
-		if (set_mime == 0) {
-			if (isbinary(buf, len - hdrlen)) {
-				error = gotweb_render_content_type_file(c,
-				    "application/octet-stream",
-				    qs->file);
-				if (error) {
-					log_warnx("%s: %s", __func__,
-					    error->msg);
-					goto done;
-				}
-			} else {
-				error = gotweb_render_content_type(c,
-				  "text/plain");
-				if (error) {
-					log_warnx("%s: %s", __func__,
-					    error->msg);
-					goto done;
-				}
-			}
-		}
-		set_mime = 1;
-		fcgi_gen_binary_response(c, buf, len - hdrlen);
+	error = got_object_blob_is_binary(binary, *blob);
+	if (error)
+		goto done;
 
-		hdrlen = 0;
-	} while (len != 0);
-done:
+ done:
 	if (commit)
 		got_object_commit_close(commit);
-	if (fd != -1 && close(fd) == -1 && error == NULL)
-		error = got_error_from_errno("close");
-	if (blob)
-		got_object_blob_close(blob);
+
+	if (error) {
+		if (*fd != -1)
+			close(*fd);
+		if (*blob)
+			got_object_blob_close(*blob);
+		*fd = -1;
+		*blob = NULL;
+	}
+
 	free(in_repo_path);
 	free(commit_id);
 	free(path);
 	return error;
 }
 
-struct blame_line {
-	int annotated;
-	char *id_str;
-	char *committer;
-	char datebuf[11]; /* YYYY-MM-DD + NUL */
-};
+int
+got_output_blob_by_lines(struct template *tp, struct got_blob_object *blob,
+    int (*cb)(struct template *, const char *, size_t))
+{
+	const struct got_error	*err;
+	char			*line = NULL;
+	size_t			 linesize = 0;
+	size_t			 lineno = 0;
+	ssize_t			 linelen = 0;
+
+	for (;;) {
+		err = got_object_blob_getline(&line, &linelen, &linesize,
+		    blob);
+		if (err || linelen == -1)
+			break;
+		lineno++;
+		if (cb(tp, line, lineno) == -1)
+			break;
+	}
+
+	free(line);
+
+	if (err) {
+		log_warnx("%s: got_object_blob_getline failed: %s",
+		    __func__, err->msg);
+		return -1;
+	}
+	return 0;
+}
 
 struct blame_cb_args {
 	struct blame_line *lines;
@@ -1094,6 +994,7 @@ struct blame_cb_args {
 	FILE *f;
 	struct got_repository *repo;
 	struct request *c;
+	got_render_blame_line_cb cb;
 };
 
 static const struct got_error *
@@ -1104,14 +1005,11 @@ got_gotweb_blame_cb(void *arg, int nlines, int lineno,
 	struct blame_cb_args *a = arg;
 	struct blame_line *bline;
 	struct request *c = a->c;
-	struct transport *t = c->t;
-	struct repo_dir *repo_dir = t->repo_dir;
-	char *line = NULL, *eline = NULL;
+	char *line = NULL;
 	size_t linesize = 0;
 	off_t offset;
 	struct tm tm;
 	time_t committer_time;
-	int r;
 
 	if (nlines != a->nlines ||
 	    (lineno != -1 && lineno < 1) || lineno > a->nlines)
@@ -1156,76 +1054,26 @@ got_gotweb_blame_cb(void *arg, int nlines, int lineno,
 	}
 
 	while (a->lineno_cur <= a->nlines && bline->annotated) {
-		char *smallerthan, *at, *nl, *committer;
-		size_t len;
-
 		if (getline(&line, &linesize, a->f) == -1) {
 			if (ferror(a->f))
 				err = got_error_from_errno("getline");
 			break;
 		}
 
-		committer = bline->committer;
-		smallerthan = strchr(committer, '<');
-		if (smallerthan && smallerthan[1] != '\0')
-			committer = smallerthan + 1;
-		at = strchr(committer, '@');
-		if (at)
-			*at = '\0';
-		len = strlen(committer);
-		if (len >= 9)
-			committer[8] = '\0';
-
-		nl = strchr(line, '\n');
-		if (nl)
-			*nl = '\0';
-
-		err = gotweb_escape_html(&eline, line);
-		if (err)
-			goto done;
-
-		if (fcgi_printf(c, "<div class='blame_wrapper'>"
-		    "<div class='blame_number'>%.*d</div>"
-		    "<div class='blame_hash'>",
-		    a->nlines_prec, a->lineno_cur) == -1)
-			goto done;
-
-		r = gotweb_link(c, &(struct gotweb_url){
-			.action = DIFF,
-			.index_page = -1,
-			.page = -1,
-			.path = repo_dir->name,
-			.commit = bline->id_str,
-		    }, "%.8s", bline->id_str);
-		if (r == -1)
-			goto done;
-
-		r = fcgi_printf(c,
-		    "</div>"
-		    "<div class='blame_date'>%s</div>"
-		    "<div class='blame_author'>%s</div>"
-		    "<div class='blame_code'>%s</div>"
-		    "</div>",	/* .blame_wrapper */
-		    bline->datebuf,
-		    committer,
-		    eline);
-		if (r == -1)
-			goto done;
+		if (a->cb(c->tp, line, bline, a->nlines_prec,
+		    a->lineno_cur) == -1)
+			break;
 
 		a->lineno_cur++;
 		bline = &a->lines[a->lineno_cur - 1];
-
-		free(eline);
-		eline = NULL;
 	}
 done:
 	free(line);
-	free(eline);
 	return err;
 }
 
 const struct got_error *
-got_output_file_blame(struct request *c)
+got_output_file_blame(struct request *c, got_render_blame_line_cb cb)
 {
 	const struct got_error *error = NULL;
 	struct transport *t = c->t;
@@ -1245,6 +1093,7 @@ got_output_file_blame(struct request *c)
 	TAILQ_INIT(&refs);
 	bca.f = NULL;
 	bca.lines = NULL;
+	bca.cb = cb;
 
 	if (asprintf(&path, "%s%s%s", qs->folder ? qs->folder : "",
 	    qs->folder ? "/" : "", qs->file) == -1) {
@@ -1381,7 +1230,7 @@ done:
 }
 
 const struct got_error *
-got_output_repo_diff(struct request *c)
+got_open_diff_for_output(FILE **fp, int *fd, struct request *c)
 {
 	const struct got_error *error = NULL;
 	struct transport *t = c->t;
@@ -1390,12 +1239,10 @@ got_output_repo_diff(struct request *c)
 	struct got_object_id *id1 = NULL, *id2 = NULL;
 	struct got_reflist_head refs;
 	FILE *f1 = NULL, *f2 = NULL, *f3 = NULL;
-	char *label1 = NULL, *label2 = NULL, *line = NULL;
-	char *newline, *eline = NULL, *color = NULL;
 	int obj_type, fd1, fd2, fd3, fd4 = -1, fd5 = -1;
-	size_t linesize = 0;
-	ssize_t linelen;
-	int wrlen = 0;
+
+	*fp = NULL;
+	*fd = -1;
 
 	TAILQ_INIT(&refs);
 
@@ -1415,14 +1262,14 @@ got_output_repo_diff(struct request *c)
 
 	if (rc->parent_id != NULL &&
 	    strncmp(rc->parent_id, "/dev/null", 9) != 0) {
-		error = got_repo_match_object_id(&id1, &label1,
+		error = got_repo_match_object_id(&id1, NULL,
 		    rc->parent_id, GOT_OBJ_TYPE_ANY,
 		    &refs, repo);
 		if (error)
 			goto done;
 	}
 
-	error = got_repo_match_object_id(&id2, &label2, rc->commit_id,
+	error = got_repo_match_object_id(&id2, NULL, rc->commit_id,
 	    GOT_OBJ_TYPE_ANY, &refs, repo);
 	if (error)
 		goto done;
@@ -1443,17 +1290,17 @@ got_output_repo_diff(struct request *c)
 	case GOT_OBJ_TYPE_BLOB:
 		error = got_diff_objects_as_blobs(NULL, NULL, f1, f2, fd4, fd5,
 		     id1, id2, NULL, NULL, GOT_DIFF_ALGORITHM_MYERS, 3, 0, 0,
-		     repo, f3);
+		     NULL, repo, f3);
 		break;
 	case GOT_OBJ_TYPE_TREE:
 		error = got_diff_objects_as_trees(NULL, NULL, f1, f2, fd4, fd5,
 		    id1, id2, NULL, "", "",  GOT_DIFF_ALGORITHM_MYERS, 3, 0, 0,
-		    repo, f3);
+		    NULL, repo, f3);
 		break;
 	case GOT_OBJ_TYPE_COMMIT:
 		error = got_diff_objects_as_commits(NULL, NULL, f1, f2, fd4,
 		    fd5, id1, id2, NULL,  GOT_DIFF_ALGORITHM_MYERS, 3, 0, 0,
-		    repo, f3);
+		    NULL, repo, f3);
 		break;
 	default:
 		error = got_error(GOT_ERR_OBJ_TYPE);
@@ -1476,109 +1323,10 @@ got_output_repo_diff(struct request *c)
 		goto done;
 	}
 
-	while ((linelen = getline(&line, &linesize, f3)) != -1) {
-		if (strncmp(line, "-", 1) == 0) {
-			color = strdup("diff_minus");
-			if (color == NULL) {
-				error = got_error_from_errno("strdup");
-				goto done;
-			}
-		} else if (strncmp(line, "+", 1) == 0) {
-			color = strdup("diff_plus");
-			if (color == NULL) {
-				error = got_error_from_errno("strdup");
-				goto done;
-			}
-		} else if (strncmp(line, "@@", 2) == 0) {
-			color = strdup("diff_chunk_header");
-			if (color == NULL) {
-				error = got_error_from_errno("strdup");
-				goto done;
-			}
-		} else if (strncmp(line, "@@", 2) == 0) {
-			color = strdup("diff_chunk_header");
-			if (color == NULL) {
-				error = got_error_from_errno("strdup");
-				goto done;
-			}
-		} else if (strncmp(line, "commit +", 8) == 0) {
-			color = strdup("diff_meta");
-			if (color == NULL) {
-				error = got_error_from_errno("strdup");
-				goto done;
-			}
-		} else if (strncmp(line, "commit -", 8) == 0) {
-			color = strdup("diff_meta");
-			if (color == NULL) {
-				error = got_error_from_errno("strdup");
-				goto done;
-			}
-		} else if (strncmp(line, "blob +", 6) == 0) {
-			color = strdup("diff_meta");
-			if (color == NULL) {
-				error = got_error_from_errno("strdup");
-				goto done;
-			}
-		} else if (strncmp(line, "blob -", 6) == 0) {
-			color = strdup("diff_meta");
-			if (color == NULL) {
-				error = got_error_from_errno("strdup");
-				goto done;
-			}
-		} else if (strncmp(line, "file +", 6) == 0) {
-			color = strdup("diff_meta");
-			if (color == NULL) {
-				error = got_error_from_errno("strdup");
-				goto done;
-			}
-		} else if (strncmp(line, "file -", 6) == 0) {
-			color = strdup("diff_meta");
-			if (color == NULL) {
-				error = got_error_from_errno("strdup");
-				goto done;
-			}
-		} else if (strncmp(line, "from:", 5) == 0) {
-			color = strdup("diff_author");
-			if (color == NULL) {
-				error = got_error_from_errno("strdup");
-				goto done;
-			}
-		} else if (strncmp(line, "via:", 4) == 0) {
-			color = strdup("diff_author");
-			if (color == NULL) {
-				error = got_error_from_errno("strdup");
-				goto done;
-			}
-		} else if (strncmp(line, "date:", 5) == 0) {
-			color = strdup("diff_date");
-			if (color == NULL) {
-				error = got_error_from_errno("strdup");
-				goto done;
-			}
-		}
+	*fp = f3;
+	*fd = fd3;
 
-		newline = strchr(line, '\n');
-		if (newline)
-			*newline = '\0';
-
-		error = gotweb_escape_html(&eline, line);
-		if (error)
-			goto done;
-
-		fcgi_printf(c, "<div class='diff_line %s'>%s</div>\n",
-		    color ? color : "", eline);
-		free(eline);
-		eline = NULL;
-
-		if (linelen > 0)
-			wrlen = wrlen + linelen;
-		free(color);
-		color = NULL;
-	}
-	if (linelen == -1 && ferror(f3))
-		error = got_error_from_errno("getline");
 done:
-	free(color);
 	if (fd4 != -1 && close(fd4) == -1 && error == NULL)
 		error = got_error_from_errno("close");
 	if (fd5 != -1 && close(fd5) == -1 && error == NULL)
@@ -1595,17 +1343,12 @@ done:
 		if (error == NULL)
 			error = f2_err;
 	}
-	if (f3) {
-		const struct got_error *f3_err =
-		    got_gotweb_flushfile(f3, fd3);
-		if (error == NULL)
-			error = f3_err;
+	if (error && f3) {
+		got_gotweb_flushfile(f3, fd3);
+		*fp = NULL;
+		*fd = -1;
 	}
 	got_ref_list_free(&refs);
-	free(line);
-	free(eline);
-	free(label1);
-	free(label2);
 	free(id1);
 	free(id2);
 	return error;
