@@ -24,6 +24,8 @@
 #include <limits.h>
 #include <stdio.h>
 
+#include "tmpl.h"
+
 #ifdef DEBUG
 #define dprintf(x...)   do { log_debug(x); } while(0)
 #else
@@ -39,6 +41,8 @@
 
 #define GOTWEBD_USER		 "www"
 
+#define GOTWEBD_MAXDESCRSZ	 1024
+#define GOTWEBD_MAXCLONEURLSZ	 1024
 #define GOTWEBD_CACHESIZE	 1024
 #define GOTWEBD_MAXCLIENTS	 1024
 #define GOTWEBD_MAXTEXT		 511
@@ -50,7 +54,7 @@
 
 /* GOTWEB DEFAULTS */
 #define MAX_QUERYSTRING		 2048
-#define MAX_SCRIPT_NAME		 255
+#define MAX_DOCUMENT_URI	 255
 #define MAX_SERVER_NAME		 255
 
 #define GOTWEB_GIT_DIR		 ".git"
@@ -113,6 +117,11 @@
 
 #define GOTWEB_PACK_NUM_TEMPFILES     32
 
+/* Forward declaration */
+struct got_blob_object;
+struct got_tree_entry;
+struct got_reflist_head;
+
 enum imsg_type {
 	IMSG_CFG_SRV = IMSG_PROC_MAX,
 	IMSG_CFG_SOCK,
@@ -136,6 +145,13 @@ struct fcgi_record_header {
 	uint8_t		reserved;
 }__attribute__((__packed__));
 
+struct blame_line {
+	int		 annotated;
+	char		*id_str;
+	char		*committer;
+	char		 datebuf[11]; /* YYYY-MM-DD + NUL */
+};
+
 struct repo_dir {
 	char			*name;
 	char			*owner;
@@ -200,10 +216,12 @@ enum socket_priv_fds {
 	PRIV_FDS__MAX,
 };
 
+struct template;
 struct request {
 	struct socket			*sock;
 	struct server			*srv;
 	struct transport		*t;
+	struct template			*tp;
 	struct event			 ev;
 	struct event			 tmo;
 
@@ -220,8 +238,9 @@ struct request {
 
 	char				 querystring[MAX_QUERYSTRING];
 	char				 http_host[GOTWEBD_MAXTEXT];
-	char				 script_name[MAX_SCRIPT_NAME];
+	char				 document_uri[MAX_DOCUMENT_URI];
 	char				 server_name[MAX_SERVER_NAME];
+	int				 https;
 
 	uint8_t				 request_started;
 };
@@ -344,8 +363,8 @@ struct gotwebd {
 };
 
 /*
- * URL parameter for gotweb_link.  NULL values and int set to -1 are
- * implicitly ignored, and string are properly escaped.
+ * URL parameter for gotweb_render_url.  NULL values and int set to -1
+ * are implicitly ignored, and string are properly escaped.
  */
 struct gotweb_url {
 	int		 action;
@@ -369,10 +388,8 @@ struct querystring {
 	char		*folder;
 	char		*headref;
 	int		 index_page;
-	char		*index_page_str;
 	char		*path;
 	int		 page;
-	char		*page_str;
 };
 
 struct querystring_keys {
@@ -401,6 +418,7 @@ enum querystring_elements {
 enum query_actions {
 	BLAME,
 	BLOB,
+	BLOBRAW,
 	BRIEFS,
 	COMMITS,
 	DIFF,
@@ -410,11 +428,21 @@ enum query_actions {
 	TAG,
 	TAGS,
 	TREE,
+	RSS,
 	ACTIONS__MAX,
 };
 
+enum gotweb_ref_tm {
+	TM_DIFF,
+	TM_LONG,
+	TM_RFC822,
+};
+
 extern struct gotwebd	*gotwebd_env;
 
+typedef int (*got_render_blame_line_cb)(struct template *, const char *,
+    struct blame_line *, int, int);
+
 /* sockets.c */
 void sockets(struct privsep *, struct privsep_proc *);
 void sockets_shutdown(void);
@@ -424,20 +452,40 @@ int sockets_privinit(struct gotwebd *, struct socket *);
 
 /* gotweb.c */
 const struct got_error *gotweb_render_content_type(struct request *,
-    const uint8_t *);
+    const char *);
 const struct got_error
-    *gotweb_render_content_type_file(struct request *, const uint8_t *, char *);
+    *gotweb_render_content_type_file(struct request *, const char *,
+    const char *, const char *);
+void gotweb_get_navs(struct request *, struct gotweb_url *, int *,
+    struct gotweb_url *, int *);
 const struct got_error *gotweb_get_time_str(char **, time_t, int);
 const struct got_error *gotweb_init_transport(struct transport **);
-const struct got_error *gotweb_escape_html(char **, const char *);
-int gotweb_link(struct request *, struct gotweb_url *, const char *, ...)
-	__attribute__((__format__(printf, 3, 4)))
-	__attribute__((__nonnull__(3)));
+const char *gotweb_action_name(int);
+int gotweb_render_url(struct request *, struct gotweb_url *);
+int gotweb_render_absolute_url(struct request *, struct gotweb_url *);
 void gotweb_free_repo_commit(struct repo_commit *);
 void gotweb_free_repo_tag(struct repo_tag *);
 void gotweb_process_request(struct request *);
 void gotweb_free_transport(struct transport *);
 
+/* pages.tmpl */
+int	gotweb_render_header(struct template *);
+int	gotweb_render_footer(struct template *);
+int	gotweb_render_repo_table_hdr(struct template *);
+int	gotweb_render_repo_fragment(struct template *, struct repo_dir *);
+int	gotweb_render_briefs(struct template *);
+int	gotweb_render_navs(struct template *);
+int	gotweb_render_commits(struct template *);
+int	gotweb_render_blob(struct template *, struct got_blob_object *);
+int	gotweb_render_tree(struct template *);
+int	gotweb_render_tags(struct template *);
+int	gotweb_render_tag(struct template *);
+int	gotweb_render_diff(struct template *, FILE *);
+int	gotweb_render_branches(struct template *, struct got_reflist_head *);
+int	gotweb_render_summary(struct template *, struct got_reflist_head *);
+int	gotweb_render_blame(struct template *);
+int	gotweb_render_rss(struct template *);
+
 /* parse.y */
 int parse_config(const char *, struct gotwebd *);
 int cmdline_symset(char *);
@@ -448,6 +496,8 @@ void fcgi_timeout(int, short, void *);
 void fcgi_cleanup_request(struct request *);
 void fcgi_create_end_record(struct request *);
 void dump_fcgi_record(const char *, struct fcgi_record_header *);
+int fcgi_puts(struct template *, const char *);
+int fcgi_putc(struct template *, int);
 int fcgi_vprintf(struct request *, const char *, va_list);
 int fcgi_printf(struct request *, const char *, ...)
 	__attribute__((__format__(printf, 2, 3)))
@@ -455,16 +505,23 @@ int fcgi_printf(struct request *, const char *, ...)
 int fcgi_gen_binary_response(struct request *, const uint8_t *, int);
 
 /* got_operations.c */
-const struct got_error *got_get_repo_owner(char **, struct request *, char *);
-const struct got_error *got_get_repo_age(char **, struct request *, char *,
+const struct got_error *got_gotweb_flushfile(FILE *, int);
+const struct got_error *got_get_repo_owner(char **, struct request *);
+const struct got_error *got_get_repo_age(char **, struct request *,
     const char *, int);
 const struct got_error *got_get_repo_commits(struct request *, int);
 const struct got_error *got_get_repo_tags(struct request *, int);
 const struct got_error *got_get_repo_heads(struct request *);
-const struct got_error *got_output_repo_diff(struct request *);
-const struct got_error *got_output_repo_tree(struct request *);
-const struct got_error *got_output_file_blob(struct request *);
-const struct got_error *got_output_file_blame(struct request *);
+const struct got_error *got_open_diff_for_output(FILE **, int *,
+    struct request *);
+int got_output_repo_tree(struct request *,
+    int (*)(struct template *, struct got_tree_entry *));
+const struct got_error *got_open_blob_for_output(struct got_blob_object **,
+    int *, int *, struct request *);
+int got_output_blob_by_lines(struct template *, struct got_blob_object *,
+    int (*)(struct template *, const char *, size_t));
+const struct got_error *got_output_file_blame(struct request *,
+    got_render_blame_line_cb);
 
 /* config.c */
 int config_setserver(struct gotwebd *, struct server *);
