@@ -39,6 +39,10 @@
 #include "got_lib_inflate.h"
 #include "got_lib_object.h"
 
+#ifndef MAX
+#define	MAX(_a,_b) ((_a) > (_b) ? (_a) : (_b))
+#endif
+
 static const struct got_error *
 add_line_metadata(struct got_diff_line **lines, size_t *nlines,
     off_t off, uint8_t type)
@@ -56,20 +60,89 @@ add_line_metadata(struct got_diff_line **lines, size_t *nlines,
 	return NULL;
 }
 
+static void
+diffstat_field_width(size_t *maxlen, int *add_cols, int *rm_cols, size_t len,
+    uint32_t add, uint32_t rm)
+{
+	int d1 = 1, d2 = 1;
+
+	if (maxlen)
+		*maxlen = MAX(*maxlen, len);
+
+	while (add /= 10)
+		++d1;
+	*add_cols = MAX(*add_cols, d1);
+
+	while (rm /= 10)
+		++d2;
+	*rm_cols = MAX(*rm_cols, d2);
+}
+
+static const struct got_error *
+get_diffstat(struct got_diffstat_cb_arg *ds, const char *path,
+    struct diff_result *r, int force_text, int status)
+{
+	const struct got_error *err;
+	struct got_pathlist_entry *pe;
+	struct got_diff_changed_path *change = NULL;
+	int flags = (r->left->atomizer_flags | r->right->atomizer_flags);
+	int isbin = (flags & DIFF_ATOMIZER_FOUND_BINARY_DATA);
+	int i;
+
+	change = calloc(1, sizeof(*change));
+	if (change == NULL)
+		return got_error_from_errno("calloc");
+
+	if (!isbin || force_text) {
+		for (i = 0; i < r->chunks.len; ++i) {
+			struct diff_chunk *c;
+			int clc, crc;
+
+			c = diff_chunk_get(r, i);
+			clc = diff_chunk_get_left_count(c);
+			crc = diff_chunk_get_right_count(c);
+
+			if (crc && !clc)
+				change->add += crc;
+			if (clc && !crc)
+				change->rm += clc;
+		}
+	}
+
+	change->status = status;
+	ds->ins += change->add;
+	ds->del += change->rm;
+	++ds->nfiles;
+
+	err = got_pathlist_append(ds->paths, path, change);
+	if (err) {
+		free(change);
+		return err;
+	}
+
+	pe = TAILQ_LAST(ds->paths, got_pathlist_head);
+	diffstat_field_width(&ds->max_path_len, &ds->add_cols, &ds->rm_cols,
+	    pe->path_len, change->add, change->rm);
+
+	return NULL;
+}
+
 static const struct got_error *
 diff_blobs(struct got_diff_line **lines, size_t *nlines,
     struct got_diffreg_result **resultp, struct got_blob_object *blob1,
     struct got_blob_object *blob2, FILE *f1, FILE *f2,
     const char *label1, const char *label2, mode_t mode1, mode_t mode2,
-    int diff_context, int ignore_whitespace, int force_text_diff, FILE *outfile,
+    int diff_context, int ignore_whitespace, int force_text_diff,
+    struct got_diffstat_cb_arg *diffstat, FILE *outfile,
     enum got_diff_algorithm diff_algo)
 {
 	const struct got_error *err = NULL, *free_err;
 	char hex1[SHA1_DIGEST_STRING_LENGTH];
 	char hex2[SHA1_DIGEST_STRING_LENGTH];
 	const char *idstr1 = NULL, *idstr2 = NULL;
+	char *modestr1 = NULL, *modestr2 = NULL;
 	off_t size1, size2;
-	struct got_diffreg_result *result;
+	struct got_diffreg_result *result = NULL;
 	off_t outoff = 0;
 	int n;
 
@@ -116,8 +189,8 @@ diff_blobs(struct got_diff_line **lines, size_t *nlines,
 		idstr2 = "/dev/null";
 
 	if (outfile) {
-		char *modestr1 = NULL, *modestr2 = NULL;
 		int modebits;
+
 		if (mode1 && mode1 != mode2) {
 			if (S_ISLNK(mode1))
 				modebits = S_IFLNK;
@@ -163,15 +236,55 @@ diff_blobs(struct got_diff_line **lines, size_t *nlines,
 			if (err)
 				goto done;
 		}
-
-		free(modestr1);
-		free(modestr2);
 	}
+
 	err = got_diffreg(&result, f1, f2, diff_algo, ignore_whitespace,
 	     force_text_diff);
 	if (err)
 		goto done;
 
+	if (diffstat) {
+		char	*path = NULL;
+		int	 status = GOT_STATUS_NO_CHANGE;
+
+		if (blob1 == NULL)
+			status = GOT_STATUS_ADD;
+		else if (blob2 == NULL)
+			status = GOT_STATUS_DELETE;
+		else {
+			if (strcmp(idstr1, idstr2) != 0)
+				status = GOT_STATUS_MODIFY;
+			else if (mode1 != mode2)
+				status = GOT_STATUS_MODE_CHANGE;
+		}
+
+		if (label1 == NULL && label2 == NULL) {
+			/* diffstat of blobs, show hash instead of path */
+			if (asprintf(&path, "%.10s -> %.10s",
+			    idstr1, idstr2) == -1) {
+				err = got_error_from_errno("asprintf");
+				goto done;
+			}
+		} else {
+			if (label2 != NULL &&
+			    (status != GOT_STATUS_DELETE || label1 == NULL))
+				path = strdup(label2);
+			else
+				path = strdup(label1);
+			if (path == NULL) {
+				err = got_error_from_errno("strdup");
+				goto done;
+			}
+		}
+
+		err = get_diffstat(diffstat, path, result->result,
+		    force_text_diff, status);
+		if (err) {
+			free(path);
+			goto done;
+		}
+	}
+
 	if (outfile) {
 		err = got_diffreg_output(lines, nlines, result,
 		    blob1 != NULL, blob2 != NULL,
@@ -182,14 +295,17 @@ diff_blobs(struct got_diff_line **lines, size_t *nlines,
 			goto done;
 	}
 
+done:
+	free(modestr1);
+	free(modestr2);
 	if (resultp && err == NULL)
 		*resultp = result;
-	else {
+	else if (result) {
 		free_err = got_diffreg_result_free(result);
 		if (free_err && err == NULL)
 			err = free_err;
 	}
-done:
+
 	return err;
 }
 
@@ -204,7 +320,8 @@ got_diff_blob_output_unidiff(void *arg, struct got_blob_object *blob1,
 
 	return diff_blobs(&a->lines, &a->nlines, NULL,
 	    blob1, blob2, f1, f2, label1, label2, mode1, mode2, a->diff_context,
-	    a->ignore_whitespace, a->force_text_diff, a->outfile, a->diff_algo);
+	    a->ignore_whitespace, a->force_text_diff, a->diffstat, a->outfile,
+	    a->diff_algo);
 }
 
 const struct got_error *
@@ -212,11 +329,12 @@ got_diff_blob(struct got_diff_line **lines, size_t*nlines,
     struct got_blob_object *blob1, struct got_blob_object *blob2,
     FILE *f1, FILE *f2, const char *label1, const char *label2,
     enum got_diff_algorithm diff_algo, int diff_context,
-    int ignore_whitespace, int force_text_diff, FILE *outfile)
+    int ignore_whitespace, int force_text_diff,
+    struct got_diffstat_cb_arg *ds, FILE *outfile)
 {
 	return diff_blobs(lines, nlines, NULL, blob1, blob2, f1, f2,
 	    label1, label2, 0, 0, diff_context, ignore_whitespace,
-	    force_text_diff, outfile, diff_algo);
+	    force_text_diff, ds, outfile, diff_algo);
 }
 
 static const struct got_error *
@@ -224,7 +342,7 @@ diff_blob_file(struct got_diffreg_result **resultp,
     struct got_blob_object *blob1, FILE *f1, off_t size1, const char *label1,
     FILE *f2, int f2_exists, struct stat *sb2, const char *label2,
     enum got_diff_algorithm diff_algo, int diff_context, int ignore_whitespace,
-    int force_text_diff, FILE *outfile)
+    int force_text_diff, struct got_diffstat_cb_arg *diffstat, FILE *outfile)
 {
 	const struct got_error *err = NULL, *free_err;
 	char hex1[SHA1_DIGEST_STRING_LENGTH];
@@ -273,6 +391,41 @@ diff_blob_file(struct got_diffreg_result **resultp,
 			goto done;
 	}
 
+	if (diffstat) {
+		char	*path = NULL;
+		int	 status = GOT_STATUS_NO_CHANGE;
+
+		/*
+		 * Ignore 'm'ode status change: if there's no accompanying
+		 * content change, there'll be no diffstat, and if there
+		 * are actual changes, 'M'odified takes precedence.
+		 */
+		if (blob1 == NULL)
+			status = GOT_STATUS_ADD;
+		else if (!f2_exists)
+			status = GOT_STATUS_DELETE;
+		else
+			status = GOT_STATUS_MODIFY;
+
+		if (label2 != NULL &&
+		    (status != GOT_STATUS_DELETE || label1 == NULL))
+			path = strdup(label2);
+		else
+			path = strdup(label1);
+		if (path == NULL) {
+			err = got_error_from_errno("strdup");
+			goto done;
+		}
+
+		err = get_diffstat(diffstat, path, result->result,
+		    force_text_diff, status);
+		if (err) {
+			free(path);
+			goto done;
+		}
+	}
+
+done:
 	if (resultp && err == NULL)
 		*resultp = result;
 	else if (result) {
@@ -280,7 +433,6 @@ diff_blob_file(struct got_diffreg_result **resultp,
 		if (free_err && err == NULL)
 			err = free_err;
 	}
-done:
 	return err;
 }
 
@@ -288,11 +440,12 @@ const struct got_error *
 got_diff_blob_file(struct got_blob_object *blob1, FILE *f1, off_t size1,
     const char *label1, FILE *f2, int f2_exists, struct stat *sb2,
     const char *label2, enum got_diff_algorithm diff_algo, int diff_context,
-    int ignore_whitespace, int force_text_diff, FILE *outfile)
+    int ignore_whitespace, int force_text_diff,
+    struct got_diffstat_cb_arg *ds, FILE *outfile)
 {
 	return diff_blob_file(NULL, blob1, f1, size1, label1, f2, f2_exists,
 	    sb2, label2, diff_algo, diff_context, ignore_whitespace,
-	    force_text_diff, outfile);
+	    force_text_diff, ds, outfile);
 }
 
 static const struct got_error *
@@ -607,6 +760,78 @@ diff_entry_new_old(struct got_tree_entry *te2,
 	    NULL, label2, 0, te2->mode, repo);
 }
 
+const struct got_error *
+got_diff_tree_compute_diffstat(void *arg, struct got_blob_object *blob1,
+    struct got_blob_object *blob2, FILE *f1, FILE *f2,
+    struct got_object_id *id1, struct got_object_id *id2,
+    const char *label1, const char *label2,
+    mode_t mode1, mode_t mode2, struct got_repository *repo)
+{
+	const struct got_error		*err = NULL;
+	struct got_diffreg_result	*result = NULL;
+	struct got_diffstat_cb_arg	*a = arg;
+	char				*path = NULL;
+	int				 status = GOT_STATUS_NO_CHANGE;
+
+	path = strdup(label2 ? label2 : label1);
+	if (path == NULL)
+		return got_error_from_errno("strdup");
+
+	if (id1 == NULL)
+		status = GOT_STATUS_ADD;
+	else if (id2 == NULL)
+		status = GOT_STATUS_DELETE;
+	else {
+		if (got_object_id_cmp(id1, id2) != 0)
+			status = GOT_STATUS_MODIFY;
+		else if (mode1 != mode2)
+			status = GOT_STATUS_MODE_CHANGE;
+	}
+
+	if (f1) {
+		err = got_opentemp_truncate(f1);
+		if (err)
+			goto done;
+	}
+	if (f2) {
+		err = got_opentemp_truncate(f2);
+		if (err)
+			goto done;
+	}
+
+	if (blob1) {
+		err = got_object_blob_dump_to_file(NULL, NULL, NULL, f1,
+		    blob1);
+		if (err)
+			goto done;
+	}
+	if (blob2) {
+		err = got_object_blob_dump_to_file(NULL, NULL, NULL, f2,
+		    blob2);
+		if (err)
+			goto done;
+	}
+
+	err = got_diffreg(&result, f1, f2, a->diff_algo, a->ignore_ws,
+	    a->force_text);
+	if (err)
+		goto done;
+
+	err = get_diffstat(a, path, result->result, a->force_text, status);
+
+done:
+	if (result) {
+		const struct got_error *free_err;
+
+		free_err = got_diffreg_result_free(result);
+		if (free_err && err == NULL)
+			err = free_err;
+	}
+	if (err)
+		free(path);
+	return err;
+}
+
 const struct got_error *
 got_diff_tree_collect_changed_paths(void *arg, struct got_blob_object *blob1,
     struct got_blob_object *blob2, FILE *f1, FILE *f2,
@@ -621,7 +846,7 @@ got_diff_tree_collect_changed_paths(void *arg, struct got_blob_object *blob1,
 
 	path = strdup(label2 ? label2 : label1);
 	if (path == NULL)
-		return got_error_from_errno("malloc");
+		return got_error_from_errno("strdup");
 
 	change = malloc(sizeof(*change));
 	if (change == NULL) {
@@ -672,13 +897,16 @@ got_diff_tree(struct got_tree_object *tree1, struct got_tree_object *tree2,
 	if (tree2) {
 		te2 = got_object_tree_get_entry(tree2, 0);
 		if (te2 && asprintf(&l2, "%s%s%s", label2, label2[0] ? "/" : "",
-		    te2->name) == -1)
-			return got_error_from_errno("asprintf");
+		    te2->name) == -1) {
+			err = got_error_from_errno("asprintf");
+			goto done;
+		}
 	}
 
 	do {
 		if (te1) {
 			struct got_tree_entry *te = NULL;
+
 			if (tree2)
 				te = got_object_tree_find_entry(tree2,
 				    te1->name);
@@ -686,10 +914,12 @@ got_diff_tree(struct got_tree_object *tree1, struct got_tree_object *tree2,
 				free(l2);
 				l2 = NULL;
 				if (te && asprintf(&l2, "%s%s%s", label2,
-				    label2[0] ? "/" : "", te->name) == -1)
-					return
-					    got_error_from_errno("asprintf");
+				    label2[0] ? "/" : "", te->name) == -1) {
+					err = got_error_from_errno("asprintf");
+					goto done;
+				}
 			}
+
 			err = diff_entry_old_new(te1, te, f1, f2, fd1, fd2,
 			    l1, l2, repo, cb, cb_arg, diff_content);
 			if (err)
@@ -698,21 +928,27 @@ got_diff_tree(struct got_tree_object *tree1, struct got_tree_object *tree2,
 
 		if (te2) {
 			struct got_tree_entry *te = NULL;
+
 			if (tree1)
 				te = got_object_tree_find_entry(tree1,
 				    te2->name);
+
 			free(l2);
+			l2 = NULL;
 			if (te) {
 				if (asprintf(&l2, "%s%s%s", label2,
-				    label2[0] ? "/" : "", te->name) == -1)
-					return
-					    got_error_from_errno("asprintf");
+				    label2[0] ? "/" : "", te->name) == -1) {
+					err = got_error_from_errno("asprintf");
+					goto done;
+				}
 			} else {
 				if (asprintf(&l2, "%s%s%s", label2,
-				    label2[0] ? "/" : "", te2->name) == -1)
-					return
-					    got_error_from_errno("asprintf");
+				    label2[0] ? "/" : "", te2->name) == -1) {
+					err = got_error_from_errno("asprintf");
+					goto done;
+				}
 			}
+
 			err = diff_entry_new_old(te2, te, f1, f2, fd2, l2,
 			    repo, cb, cb_arg, diff_content);
 			if (err)
@@ -726,9 +962,12 @@ got_diff_tree(struct got_tree_object *tree1, struct got_tree_object *tree2,
 			te1 = got_object_tree_get_entry(tree1, tidx1);
 			if (te1 &&
 			    asprintf(&l1, "%s%s%s", label1,
-			    label1[0] ? "/" : "", te1->name) == -1)
-				return got_error_from_errno("asprintf");
+			    label1[0] ? "/" : "", te1->name) == -1) {
+				err = got_error_from_errno("asprintf");
+				goto done;
+			}
 		}
+
 		free(l2);
 		l2 = NULL;
 		if (te2) {
@@ -736,11 +975,16 @@ got_diff_tree(struct got_tree_object *tree1, struct got_tree_object *tree2,
 			te2 = got_object_tree_get_entry(tree2, tidx2);
 			if (te2 &&
 			    asprintf(&l2, "%s%s%s", label2,
-			        label2[0] ? "/" : "", te2->name) == -1)
-				return got_error_from_errno("asprintf");
+			    label2[0] ? "/" : "", te2->name) == -1) {
+				err = got_error_from_errno("asprintf");
+				goto done;
+			}
 		}
 	} while (te1 || te2);
 
+done:
+	free(l1);
+	free(l2);
 	return err;
 }
 
@@ -750,7 +994,7 @@ got_diff_objects_as_blobs(struct got_diff_line **lines, size_t *nlines,
     struct got_object_id *id1, struct got_object_id *id2,
     const char *label1, const char *label2,
     enum got_diff_algorithm diff_algo, int diff_context,
-    int ignore_whitespace, int force_text_diff,
+    int ignore_whitespace, int force_text_diff, struct got_diffstat_cb_arg *ds,
     struct got_repository *repo, FILE *outfile)
 {
 	const struct got_error *err;
@@ -771,7 +1015,7 @@ got_diff_objects_as_blobs(struct got_diff_line **lines, size_t *nlines,
 	}
 	err = got_diff_blob(lines, nlines, blob1, blob2, f1, f2, label1, label2,
 	    diff_algo, diff_context, ignore_whitespace, force_text_diff,
-	    outfile);
+	    ds, outfile);
 done:
 	if (blob1)
 		got_object_blob_close(blob1);
@@ -945,8 +1189,8 @@ diff_objects_as_trees(struct got_diff_line **lines, size_t *nlines,
     struct got_object_id *id1, struct got_object_id *id2,
     struct got_pathlist_head *paths, const char *label1, const char *label2,
     int diff_context, int ignore_whitespace, int force_text_diff,
-    struct got_repository *repo, FILE *outfile,
-    enum got_diff_algorithm diff_algo)
+    struct got_diffstat_cb_arg *dsa, struct got_repository *repo,
+    FILE *outfile, enum got_diff_algorithm diff_algo)
 {
 	const struct got_error *err;
 	struct got_tree_object *tree1 = NULL, *tree2 = NULL;
@@ -971,6 +1215,7 @@ diff_objects_as_trees(struct got_diff_line **lines, size_t *nlines,
 	arg.diff_context = diff_context;
 	arg.ignore_whitespace = ignore_whitespace;
 	arg.force_text_diff = force_text_diff;
+	arg.diffstat = dsa;
 	arg.outfile = outfile;
 	if (want_linemeta) {
 		arg.lines = *lines;
@@ -979,14 +1224,12 @@ diff_objects_as_trees(struct got_diff_line **lines, size_t *nlines,
 		arg.lines = NULL;
 		arg.nlines = 0;
 	}
-	if (paths == NULL || TAILQ_EMPTY(paths)) {
-		err = got_diff_tree(tree1, tree2, f1, f2, fd1, fd2,
-		    label1, label2, repo,
-		    got_diff_blob_output_unidiff, &arg, 1);
-	} else {
+	if (paths == NULL || TAILQ_EMPTY(paths))
+		err = got_diff_tree(tree1, tree2, f1, f2, fd1, fd2, label1,
+		    label2, repo, got_diff_blob_output_unidiff, &arg, 1);
+	else
 		err = diff_paths(tree1, tree2, f1, f2, fd1, fd2, paths, repo,
 		    got_diff_blob_output_unidiff, &arg);
-	}
 	if (want_linemeta) {
 		*lines = arg.lines; /* was likely re-allocated */
 		*nlines = arg.nlines;
@@ -1005,7 +1248,8 @@ got_diff_objects_as_trees(struct got_diff_line **lines, size_t *nlines,
     struct got_object_id *id1, struct got_object_id *id2,
     struct got_pathlist_head *paths, const char *label1, const char *label2,
     enum got_diff_algorithm diff_algo, int diff_context, int ignore_whitespace,
-    int force_text_diff, struct got_repository *repo, FILE *outfile)
+    int force_text_diff, struct got_diffstat_cb_arg *dsa,
+    struct got_repository *repo, FILE *outfile)
 {
 	const struct got_error *err;
 	char *idstr = NULL;
@@ -1047,7 +1291,7 @@ got_diff_objects_as_trees(struct got_diff_line **lines, size_t *nlines,
 
 	err = diff_objects_as_trees(lines, nlines, f1, f2, fd1, fd2, id1, id2,
 	    paths, label1, label2, diff_context, ignore_whitespace,
-	    force_text_diff, repo, outfile, diff_algo);
+	    force_text_diff, dsa, repo, outfile, diff_algo);
 done:
 	free(idstr);
 	return err;
@@ -1059,7 +1303,7 @@ got_diff_objects_as_commits(struct got_diff_line **lines, size_t *nlines,
     struct got_object_id *id1, struct got_object_id *id2,
     struct got_pathlist_head *paths, enum got_diff_algorithm diff_algo,
     int diff_context, int ignore_whitespace, int force_text_diff,
-    struct got_repository *repo, FILE *outfile)
+    struct got_diffstat_cb_arg *dsa, struct got_repository *repo, FILE *outfile)
 {
 	const struct got_error *err;
 	struct got_commit_object *commit1 = NULL, *commit2 = NULL;
@@ -1102,8 +1346,8 @@ got_diff_objects_as_commits(struct got_diff_line **lines, size_t *nlines,
 	err = diff_objects_as_trees(lines, nlines, f1, f2, fd1, fd2,
 	    commit1 ? got_object_commit_get_tree_id(commit1) : NULL,
 	    got_object_commit_get_tree_id(commit2), paths, "", "",
-	    diff_context, ignore_whitespace, force_text_diff, repo, outfile,
-	    diff_algo);
+	    diff_context, ignore_whitespace, force_text_diff, dsa, repo,
+	    outfile, diff_algo);
 done:
 	if (commit1)
 		got_object_commit_close(commit1);
@@ -1150,6 +1394,7 @@ done:
 		*resultp = diffreg_result;
 	else if (diffreg_result) {
 		const struct got_error *free_err;
+
 		free_err = got_diffreg_result_free(diffreg_result);
 		if (free_err && err == NULL)
 			err = free_err;
