@@ -401,6 +401,39 @@ got_object_blob_read_block(size_t *outlenp, struct got_blob_object *blob)
 	return NULL;
 }
 
+const struct got_error *
+got_object_blob_is_binary(int *binary, struct got_blob_object *blob)
+{
+	const struct got_error *err;
+	size_t hdrlen, len;
+
+	*binary = 0;
+	hdrlen = got_object_blob_get_hdrlen(blob);
+
+	if (fseeko(blob->f, hdrlen, SEEK_SET) == -1)
+		return got_error_from_errno("fseeko");
+
+	err = got_object_blob_read_block(&len, blob);
+	if (err)
+		return err;
+
+	*binary = memchr(blob->read_buf, '\0', len) != NULL;
+
+	if (fseeko(blob->f, hdrlen, SEEK_SET) == -1)
+		return got_error_from_errno("fseeko");
+	return NULL;
+}
+
+const struct got_error *
+got_object_blob_getline(char **line, ssize_t *linelen, size_t *linesize,
+    struct got_blob_object *blob)
+{
+	*linelen = getline(line, linesize, blob->f);
+	if (*linelen == -1 && !feof(blob->f))
+		return got_error_from_errno("getline");
+	return NULL;
+}
+
 const struct got_error *
 got_object_blob_dump_to_file(off_t *filesize, int *nlines,
     off_t **line_offsets, FILE *outfile, struct got_blob_object *blob)
@@ -923,7 +956,7 @@ got_object_commit_retain(struct got_commit_object *commit)
 
 const struct got_error *
 got_object_raw_alloc(struct got_raw_object **obj, uint8_t *outbuf, int *outfd,
-    size_t hdrlen, off_t size)
+    size_t max_in_mem_size, size_t hdrlen, off_t size)
 {
 	const struct got_error *err = NULL;
 	off_t tot;
@@ -948,11 +981,12 @@ got_object_raw_alloc(struct got_raw_object **obj, uint8_t *outbuf, int *outfd,
 		}
 
 		if (sb.st_size != tot) {
-			err = got_error(GOT_ERR_PRIVSEP_LEN);
+			err = got_error_msg(GOT_ERR_BAD_OBJ_HDR,
+			    "raw object has unexpected size");
 			goto done;
 		}
 #ifndef GOT_PACK_NO_MMAP
-		if (tot > 0 && tot <= SIZE_MAX) {
+		if (tot > 0 && tot <= max_in_mem_size) {
 			(*obj)->data = mmap(NULL, tot, PROT_READ,
 			    MAP_PRIVATE, *outfd, 0);
 			if ((*obj)->data == MAP_FAILED) {
