@@ -409,12 +409,14 @@ parse_gmtoff(time_t *gmtoff, const char *tzstr)
 	else if (*p != '+')
 		return got_error(GOT_ERR_BAD_OBJ_DATA);
 	p++;
-	if (!isdigit(*p) && !isdigit(*(p + 1)))
+	if (!isdigit((unsigned char)*p) &&
+	    !isdigit((unsigned char)*(p + 1)))
 		return got_error(GOT_ERR_BAD_OBJ_DATA);
 	h = (((*p - '0') * 10) + (*(p + 1) - '0'));
 
 	p += 2;
-	if (!isdigit(*p) && !isdigit(*(p + 1)))
+	if (!isdigit((unsigned char)*p) &&
+	    !isdigit((unsigned char)*(p + 1)))
 		return got_error(GOT_ERR_BAD_OBJ_DATA);
 	m = ((*p - '0') * 10) + (*(p + 1) - '0');
 
@@ -821,7 +823,7 @@ parse_tree_entry(struct got_parsed_tree_entry *pte, size_t *elen, char *buf,
 	pte->mode = 0;
 	p = buf;
 	while (p < space) {
-		if (*p < '0' && *p > '7')
+		if (*p < '0' || *p > '7')
 			return got_error(GOT_ERR_BAD_OBJ_DATA);
 		pte->mode <<= 3;
 		pte->mode |= *p - '0';
