@@ -265,8 +265,7 @@ echo_error(const struct got_error *err, int outfd, int chattygot)
 	 */
 	buf[0] = 'E'; buf[1] = 'R'; buf[2] = 'R'; buf[3] = ' '; buf[4] = '\0';
 	len = strlcat(buf, err->msg, sizeof(buf));
-	err = got_pkt_writepkt(outfd, buf, len, chattygot);
-	abort();
+	got_pkt_writepkt(outfd, buf, len, chattygot);
 }
 
 static const struct got_error *
@@ -340,14 +339,12 @@ announce_refs(int outfd, struct imsgbuf *ibuf, int client_is_reading,
 				err = got_error(GOT_ERR_PRIVSEP_LEN);
 				goto done;
 			}
-			refname = malloc(iref.name_len + 1);
+			refname = strndup(imsg.data + sizeof(iref),
+			    iref.name_len);
 			if (refname == NULL) {
-				err = got_error_from_errno("malloc");
+				err = got_error_from_errno("strndup");
 				goto done;
 			}
-			memcpy(refname, imsg.data + sizeof(iref),
-			    iref.name_len);
-			refname[iref.name_len] = '\0';
 			err = send_ref(outfd, iref.id, refname,
 			    !sent_capabilities, client_is_reading,
 			    NULL, chattygot);
@@ -383,24 +380,20 @@ announce_refs(int outfd, struct imsgbuf *ibuf, int client_is_reading,
 			    symrefname != NULL || symreftarget != NULL)
 				break;
 
-			symrefname = malloc(isymref.name_len + 1);
+			symrefname = strndup(imsg.data + sizeof(isymref),
+			    isymref.name_len);
 			if (symrefname == NULL) {
 				err = got_error_from_errno("malloc");
 				goto done;
 			}
-			memcpy(symrefname, imsg.data + sizeof(isymref),
-			    isymref.name_len);
-			symrefname[isymref.name_len] = '\0';
 
-			symreftarget = malloc(isymref.target_len + 1);
-			if (symreftarget == NULL) {
-				err = got_error_from_errno("malloc");
-				goto done;
-			}
-			memcpy(symreftarget,
+			symreftarget = strndup(
 			    imsg.data + sizeof(isymref) + isymref.name_len,
 			    isymref.target_len);
-			symreftarget[isymref.target_len] = '\0';
+			if (symreftarget == NULL) {
+				err = got_error_from_errno("strndup");
+				goto done;
+			}
 
 			if (asprintf(&symrefstr, "%s:%s", symrefname,
 			    symreftarget) == -1) {
@@ -905,12 +898,36 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 		if (err)
 			break;
 		if (n == 0) {
-			if (curstate != STATE_EXPECT_MORE_WANT &&
-			    curstate != STATE_EXPECT_HAVE) {
+			if (curstate != STATE_EXPECT_WANT &&
+			    curstate != STATE_EXPECT_MORE_WANT &&
+			    curstate != STATE_EXPECT_HAVE &&
+			    curstate != STATE_EXPECT_DONE) {
 				err = got_error_msg(GOT_ERR_BAD_PACKET,
 				    "unexpected flush packet received");
 				goto done;
 			}
+
+			if (curstate == STATE_EXPECT_WANT) {
+				ssize_t r;
+				/*
+				 * If the client does not want to fetch
+				 * anything we should receive a flush
+				 * packet followed by EOF.
+				 */
+				r = read(infd, buf, sizeof(buf));
+				if (r == -1) {
+					err = got_error_from_errno("read");
+					goto done;
+				}
+				if (r == 0) /* EOF */
+					goto done;
+
+				/* Zero-length field followed by payload. */
+				err = got_error_msg(GOT_ERR_BAD_PACKET,
+				    "unexpected flush packet received");
+				goto done;
+			}
+
 			err = forward_flushpkt(&ibuf);
 			if (err)
 				goto done;
@@ -937,16 +954,21 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 			if (curstate == STATE_EXPECT_WANT)
 				curstate = STATE_EXPECT_MORE_WANT;
 		} else if (n >= 5 && strncmp(buf, "have ", 5) == 0) {
-			if (curstate != STATE_EXPECT_HAVE) {
+			if (curstate != STATE_EXPECT_HAVE &&
+			    curstate != STATE_EXPECT_DONE) {
 				err = got_error_msg(GOT_ERR_BAD_PACKET,
 				    "unexpected 'have' packet");
 				goto done;
 			}
-			err = recv_have(&have_ack, outfd, &ibuf, buf, n,
-			    chattygot);
-			if (err)
-				goto done;
-			seen_have = 1;
+			if (curstate == STATE_EXPECT_HAVE) {
+				err = recv_have(&have_ack, outfd, &ibuf,
+				    buf, n, chattygot);
+				if (err)
+					goto done;
+				seen_have = 1;
+				if (have_ack)
+					curstate = STATE_EXPECT_DONE;
+			}
 		} else if (n == 5 && strncmp(buf, "done\n", 5) == 0) {
 			if (curstate != STATE_EXPECT_HAVE &&
 			    curstate != STATE_EXPECT_DONE) {
@@ -1220,13 +1242,11 @@ report_unpack_status(struct imsg *imsg, int outfd, int chattygot)
 	if (datalen != sizeof(istatus) + istatus.reason_len)
 		return got_error(GOT_ERR_PRIVSEP_LEN);
 
-	reason = malloc(istatus.reason_len + 1);
+	reason = strndup(imsg->data + sizeof(istatus), istatus.reason_len);
 	if (reason == NULL) {
-		err = got_error_from_errno("malloc");
+		err = got_error_from_errno("strndup");
 		goto done;
 	}
-	memcpy(reason, imsg->data + sizeof(istatus), istatus.reason_len);
-	reason[istatus.reason_len] = '\0';
 
 	if (err == NULL)
 		len = snprintf(buf, sizeof(buf), "unpack ok\n");
@@ -1261,11 +1281,9 @@ recv_ref_update_ok(struct imsg *imsg, int outfd, int chattygot)
 
 	memcpy(&iok, imsg->data, sizeof(iok));
 
-	refname = malloc(iok.name_len + 1);
+	refname = strndup(imsg->data + sizeof(iok), iok.name_len);
 	if (refname == NULL)
-		return got_error_from_errno("malloc");
-	memcpy(refname, imsg->data + sizeof(iok), iok.name_len);
-	refname[iok.name_len] = '\0';
+		return got_error_from_errno("strndup");
 
 	len = snprintf(buf, sizeof(buf), "ok %s\n", refname);
 	if (len >= sizeof(buf)) {
@@ -1297,20 +1315,16 @@ recv_ref_update_ng(struct imsg *imsg, int outfd, int chattygot)
 
 	memcpy(&ing, imsg->data, sizeof(ing));
 
-	refname = malloc(ing.name_len + 1);
+	refname = strndup(imsg->data + sizeof(ing), ing.name_len);
 	if (refname == NULL)
-		return got_error_from_errno("malloc");
-	memcpy(refname, imsg->data + sizeof(ing), ing.name_len);
-	refname[ing.name_len] = '\0';
+		return got_error_from_errno("strndup");
 
-	reason = malloc(ing.reason_len + 1);
+	reason = strndup(imsg->data + sizeof(ing) + ing.name_len,
+	    ing.reason_len);
 	if (reason == NULL) {
-		err = got_error_from_errno("malloc");
+		err = got_error_from_errno("strndup");
 		goto done;
 	}
-	memcpy(refname, imsg->data + sizeof(ing) + ing.name_len,
-	    ing.reason_len);
-	refname[ing.reason_len] = '\0';
 
 	len = snprintf(buf, sizeof(buf), "ng %s %s\n", refname, reason);
 	if (len >= sizeof(buf)) {
